@@ -28,13 +28,17 @@
 
 #include <xwayland-config.h>
 
+#include <sys/types.h>
+
 #include <stdio.h>
 #include <unistd.h>
+
 #include <X11/X.h>
 #include <dix.h>
 #include <propertyst.h>
 #include <validate.h>
 #include <wayland-util.h>
+#include <xf86drm.h>
 
 #include "xwayland-types.h"
 
@@ -51,7 +55,7 @@ struct xwl_format_table_entry {
 };
 
 struct xwl_device_formats {
-    dev_t drm_dev;
+    drmDevice *drm_dev;
     int supports_scanout;
     uint32_t num_formats;
     struct xwl_format *formats;
@@ -71,7 +75,7 @@ struct xwl_format_table {
 struct xwl_dmabuf_feedback {
     struct zwp_linux_dmabuf_feedback_v1 *dmabuf_feedback;
     struct xwl_format_table format_table;
-    dev_t main_dev;
+    drmDevice *main_dev;
     /*
      * This will be filled in during wl events and copied to
      * dev_formats on dmabuf_feedback.tranche_done
